#define I_SYS
#define I_ERRNO
#define I_GETOPT
#define I_STRING
#define I_SOCKET
#include "includes.h"
#include "client.h"

#include <sys/wait.h>
#include <signal.h>

#include "debug.h"

/*
 *
 * main file. Calls everything else. ;)
 * 
 * basically consists of select() and then calls things based on the select().
 *
 */
/*------------------------------------------------------------------------*/
/* truely global vars. Used everywhere. */
int current_time = 0;
int do_shutdown	 = 0;
int debug = 0;
int compressing = 1;
/*------------------------------------------------------------------------*/
/* Various global variables. Note that most of these vars are local to    */
/* this file. Where possible (i.e. where it doesn't impinge efficency) 	  */
/* this is enforced */
int remote;

int clients_waiting;
struct Client clients[MAX_CLIENTS];
int num_clients = 0;
int baudrate =   2400;
int bytes_left = 1024; /* bytes availiable to send this second.. */
int fudge_flow = 0; /* should we generate periodic control-Q's */
int byte_shift = 0; /* So we can map less frequently uses section down to */
                    /* 0-32 */
int window_size = 3; /* A convenient number. 9600 baud users should probly */
				/* go to 2. */
int packet_timeout = 50; /* wait 2.5 seconds for a packet to time out */
int do_forceing = 1; /* weather or not to automatically transmit packets */
				/* if the window is full, and the out */
				/* buffer is empty. */
int write_noise = 0; /* whether we should print out all the the serial stuff */
				/* we get that we don't understand.. */

int seven_bit_in = 0;		/* Are we on a line that ignores the */
				/* top bit. */
int in_mask = 255;
int seven_bit_out = 0;
int out_mask = 255;

int stat_modem_in = 0,
  stat_modem_out = 0;
char escapes[256];
char ignores[256];
/*------------------------------------------------------------------------*/
/* module function prototypes */

void main_loop(int, int, int);
void main_init(void);
void do_link_in(void);
void do_link_out(void);
/*------------------------------------------------------------------------*/
/* main. */

void main_init(void) {
  int i;
  for (i = 0; i < MAX_CLIENTS;++i) {
    clients[i].fd = -1;
    clients[i].state = -1;
  }
}
				/* In case of sudden death, do some */
				/* minimal clean up. */
void sig_quit(int dummy) {
  set_block(0);
  set_block(1);
  terminal_restore(1);
  exit(0);
}
				/* Drop a core. */
void sig_core(int dummy) {
  set_block(0);
  set_block(1);
  terminal_restore(1);
}
				/* We ignore this signal. There was */
				/* some problem with it on linux. A */
				/* bit odd, and I can't track it down. */
				/* I suspect it is a bad shell. So we */
				/* just ignore it. */
void sig_ignore(int dummy) {
  signal(SIGALRM, sig_ignore);
}

void sig_child(int dummy) {
  int p, i;
  p = wait3(0, WNOHANG, 0);
  signal(SIGCHLD, sig_child);

  if (p < 1) return;
  DEBUG_MAIN(stderr, "%s:sig child kicked\n", term_server);
  for (i = 0; i < MAX_CLIENTS;++i) {
    if (clients[i].pid == p && !clients[i].state) {
      clients[i].state = 2;
      return;
    }
  }
  /* Hmm. child that we don't know about died!?!?! */
  if (dummy)
    dummy = 0;
}

void read_rc(void) {
  FILE *f;
  char file[200];
  int i;

  if (!getenv("HOME"))
    return;
  sprintf(file, "%s/.term/termrc", getenv("HOME"));
  f = fopen(file, "r");
  if (!f) return;
  
  while (!feof(f)) {
    fgets(file, 200, f);
          /* skip blank lines + comments */
    if (!file[0] || file[0] == '\n' || file[0] == '#')
      continue;
    /* compression on/off */
    if (!strncmp(file, "compress ", 9)) {
      if (!strncmp(file + 9, "off", 3))
	compressing = 0;
    }

    /* any special characters to be escapeing */
    else if (!strncmp(file, "escape ", 7)) {	/* If it is an escape.. */
      char *p;
      int j;
      
      i = atoi(file + 7);	/* get the number following. */
      if (i < 0 || i > 255) {	/* check for sanity. */
	fprintf(stderr, "Invalid escape %d in termrc\n", i);
	continue;
      }

      if ((p = strchr(file+7, '-')) != NULL) { /* See if this is a */
					     /* range.. */
	while (*++p == ' ');	/* skip whitespace. Note that it */
				/* automatically skips the '-'. */
	j = atoi(p);		/* if it is, then get the second number. */
	if (j < 0 || j > 255) {	/* sanity check again. */
	  fprintf(stderr, "Invalid range limit %d in termrc\n", j);
	  continue;
	}
	for (;i != j; i = (i+1) & 255) /* Ok. mark all the characters */
				/* in the range as being escaped. */
	  escapes[i] = 1;   
      }
      else escapes[i] = 1;	/* else just set the one character as */
				/* to be escaped. */
    }
    else if (!strncmp(file, "ignore ", 7)) {	/* If it is an escape.. */
      char *p;
      int j;
      
      i = atoi(file + 7);	/* get the number following. */
      if (i < 0 || i > 255) {	/* check for sanity. */
	fprintf(stderr, "Invalid ignore %d in termrc\n", i);
	continue;
      }

      if ((p = strchr(file+7, '-')) != NULL) { /* See if this is a */
					     /* range.. */
	while (*++p == ' ');	/* skip whitespace. Note that it */
				/* automatically skips the '-'. */
	j = atoi(p);		/* if it is, then get the second number. */
	if (j < 0 || j > 255) {	/* sanity check again. */
	  fprintf(stderr, "Invalid range limit %d in termrc\n", j);
	  continue;
	}
	for (;i != j; i = (i+1) & 255) /* Ok. mark all the characters */
				/* in the range as being escaped. */
	  ignores[i] = 1;   
      }
      else ignores[i] = 1;	/* else just set the one character as */
				/* to be escaped. */
    }
    /* The baudrate */
    else if (!strncmp(file, "baudrate ", 9)) {
      i = atoi(file + 9);
      if (i < 300) i = 300;
      baudrate = i;
    }
    else if (!strncmp(file, "shift ", 6)) {
      byte_shift = atoi(file + 6);
      if (byte_shift < 0 || byte_shift > 255) {
	fprintf(stderr, "Invalid shift %d\n", byte_shift);
	byte_shift = 0;
      }
    }

    else if (!strncmp(file, "window ", 7)) {
      window_size = atoi(file +7 );
      if (window_size < 1 || window_size > 14) {
	fprintf(stderr, "Invalid window size %d. Must be in"
		"[1..14]\n", window_size);
	window_size = 3;
      }
    }
    
    else if (!strncmp(file, "timeout ", 8)) {
      packet_timeout = atoi(file + 8);
      if (packet_timeout < 10 || packet_timeout > 240) {
	fprintf(stderr, "Invalid timeout value (%d). Must be in"
		"[10..240]\n", packet_timeout);
	packet_timeout = 50;
      }
    }
    
    else if (!strncmp(file, "force on", 8)) {
      do_forceing = 1;
    }
    else if (!strncmp(file, "noise on", 8)) {
      write_noise = 1;
    } 
    else if (!strncmp(file, "sevenbit", 8)) {
      extern int tok_byte_mask_in, tok_byte_width_in, 
        tok_byte_mask_out, tok_byte_width_out;
      seven_bit_in = 1;
      seven_bit_out = 1;
      in_mask = out_mask = 127;
      tok_byte_mask_in = tok_byte_mask_out = 127;
      tok_byte_width_in = tok_byte_width_out = 7;
    }
    else if (!(strncmp(file, "seven_out", 9))) {
      extern int tok_byte_mask_out, tok_byte_width_out;
      seven_bit_out = 1;
      out_mask = 127;
      tok_byte_mask_out = 127;
      tok_byte_width_out = 7;
    }
    else if (!(strncmp(file, "seven_in", 8))) {
      extern int tok_byte_mask_in, tok_byte_width_in;
      seven_bit_in = 1;
      in_mask = 127;
      tok_byte_mask_in = 127;
      tok_byte_width_in = 7;
    }

    /* Should we generate flow control characters ?? */
    else if (!strncmp(file, "flowcontrol ", 12)) {
      fudge_flow = atoi(file+12);
      if (fudge_flow < 0) fudge_flow = 0;
    } 
    else {
      fprintf(stderr, "Unrecognized line in ~/.term/termrc.\n");
      fprintf(stderr, "'%s'\n", file);
    }
  }
  fclose(f);
}

void main(int argc, char *argv[]) 
{
  extern char *optarg;
  extern int optind, opterr;
  
  int s, c, i;
  char sock_unix[109];
  char *home;
  
  /* initalize character escapeing. */
  for (i = 0; i < 256;i++) {
    ignores[i] = 0;
    if (i < 32 || i == '^')
      escapes[i] =1;
    else 
      escapes[i] = 0;
   }
  
  /* read in the termrc file. */
  read_rc();

  /* then check env variables. */
  setbuf(stderr, 0);
  if (getenv("BAUDRATE")) {
    baudrate = atoi(getenv("BAUDRATE"));
    if ( baudrate < 300)
      baudrate = 300;
  }
  /* Then check command line options */
  /*
    The code to parse the command line was written by the
    one and only Muhammad Saggaf. If you have any question
    about Linux, networking, Unix, or life in general, 
    don't ask him mate! :). Chances are he doesn't know the 
    answer.
    */
  
  opterr = 0;
  
  while ((c = getopt (argc, argv, "ac:d:f:n:ors:t:w:")) !=EOF)
    switch(c) {
    case 'a':
      {
	extern int tok_byte_mask_out, tok_byte_width_out,
	  tok_byte_mask_in, tok_byte_width_in;
	tok_byte_mask_out = tok_byte_mask_in = 127;
	tok_byte_width_out = tok_byte_width_in = 7;

	seven_bit_in = seven_bit_out = 1;
	in_mask = out_mask = 127;
	break;
      }
    case 'f': 
      fudge_flow = atoi(optarg);
      if (fudge_flow < 0) fudge_flow = 0;
      break;
    case 'o': do_forceing = 1; break;      
    case 'r': remote = 1; break;
    case 't':
      packet_timeout = atoi(optarg);
      if (packet_timeout < 10 || packet_timeout > 240) {
	fprintf(stderr, "Invalid timeout value (%d).\n", packet_timeout);
	packet_timeout = 50;
      }
      break;
    case 'w':
      window_size = atoi(optarg);
      if (window_size < 1 || window_size > 14) {
	fprintf(stderr, "Invalid window size %d\n", window_size);
	window_size = 3;
      }
      break;
    case 'd': 
      debug = atoi(optarg);
      fprintf(stderr, "Debugging = %x\n", debug);
      break;
    case 'c':
      if (!strcmp(optarg, "off"))
	compressing = 0;
      break;
    case 'n':
      if (!strcmp(optarg, "on"))
	write_noise = 1;
      else write_noise = 0;
      break;
    case 's':
      baudrate = atoi(optarg);
      if (baudrate < 300) {
	fprintf(stderr, "baudrate set too low. Reset to 300\n");
	baudrate = 300;
      }
      break;
    default:
      fprintf(stderr, "unrecognized or incomplete argument '%s'. Exiting\n", \
	      argv[optind-1]);
      exit(-1);
    }
  
  if (optind < argc)
    term_server = argv[optind];
  
  for (s = 0; s < MAX_CLIENTS;++s) {
    clients[s].fd = -1;
  }
  if (!term_server) {
    term_server = "";
  }
  
  home = getenv("TERMDIR");
  if (!home)
    home = getenv("HOME");
  if (!home)
    home = "/tmp";   
  
  sprintf(sock_unix,"%s/.term", home);
  mkdir(sock_unix, 0700);
  strcat(sock_unix, "/socket");
  strcat(sock_unix, term_server);
  
  if ((s = bind_unix(sock_unix)) < 0) {
    exit(1);
  }
  /* init modules */
  
  main_init();
  serial_init();
  compress_init();
  update_time();

  terminal_save(1);
  terminal_raw(1);
  
  signal(SIGCHLD, sig_child);
  signal(SIGHUP, sig_quit);
  signal(SIGINT, sig_quit);
  signal(SIGQUIT, sig_quit);
  signal(SIGIOT, sig_core);
  signal(SIGSEGV, sig_core);
  signal(SIGALRM, sig_ignore);

  set_nonblock(0);
  set_nonblock(1);
  set_nonblock(s);
  
  main_loop(s, 0, 1);
  
  set_block(0);
  set_block(1);
  set_block(s);

  terminal_restore(1);
}

/*-----------------------------------------------------------------------*/
void check_client(int cl, int ret) {
#if 0
  if (ret == 0) {
    /* Hmm. Has it gone away ?? */
    clients[cl].fd = -1;
    return;
  }	
#endif
  if (ret < 0 && errno == ERR_BLOCK ) {
    fprintf(stderr, "%s:EWOULDBLOCK\n", term_server);
#ifndef linux
    return; 
#endif
  }
  if (clients[cl].state == 3 ) {
    DEBUG_FP(stderr, "%s:truncating out_buff\n", term_server);
    clients[cl].out_buff.size = 0;
    return;
  }
  if (ret < 0)
    perror("client gave this");
  
  clients[cl].state = 2;
}

int new_client(void) {
  int j;
  for (j = 0; j < MAX_CLIENTS;++j)
    if (clients[j].fd < 0) break;
  DEBUG_FP(stderr, "%s: new client %d.\n", term_server, j);
  if (j == MAX_CLIENTS) return -1; /* not maximum clients */
  clients[j].in_buff.size =
    clients[j].in_buff.start = 
      clients[j].in_buff.end = 
	clients[j].out_buff.size =
	  clients[j].out_buff.start =
	    clients[j].out_buff.end = 0;
  clients[j].type = T_SMART | T_RDFILE | T_WRFILE;
  clients[j].cl_type = CL_SOCKET;
  clients[j].compress = compressing;
  clients[j].state = 1;
  clients[j].c_state = 0;
  clients[j].number = j;
  clients[j].priority = 0;
  clients[j].name[0] = 0;
  return j;
}
/*------------------------------------------------------------------------*/
/* Main loop. Hangs around waiting for things to get ready, and calls to  */
/* approroprite routines. 						  */

void main_loop(int socket, int modem_in, int modem_out) {
  struct timeval timeout;
  fd_set reads, writes, excepts;
  int i, j, empty = 1;
  
  while (!do_shutdown) {
    
/* Set up client stuff */
/* We select to read if: */
/*	The input buffer is empty */
/* We select to write if: */
/*      The output buffer is not empty */

    FD_ZERO(&reads);
    FD_ZERO(&writes);
    FD_ZERO(&excepts);
    
    if (p_in_num && empty)
      do_link_in();
    
    for (i = 0; i < MAX_CLIENTS;++i) {
      if (clients[i].fd < 0) continue;
      if (clients[i].state == 3) {
	if (!clients[i].in_buff.size && !clients[i].out_buff.size)
	  {
	    DEBUG_FP(stderr, "%s:real close %d\n", term_server , i);
#ifdef X_DEBUG
	    if (clients[i].type & 128)
	      close(clients[i].pid);
#endif
	    close(clients[i].fd);
	    clients[i].fd = -1;
	    clients[i].state = -1;
	    continue;
	  }
      }
      if (clients[i].cl_type == CL_FILE) continue;
      if (!clients[i].in_buff.size && clients[i].state == 1 &&
	  (clients[i].type & T_RDFILE)) {
	FD_SET(clients[i].fd, &reads);
      }
      if (clients[i].out_buff.size && (clients[i].type & T_WRFILE)) {
	FD_SET(clients[i].fd, &writes);
      }
      FD_SET(clients[i].fd, &excepts);
    }
    
/* Select for socket and modem */
/* We select read for socket if we aren't at the maxmimum number */
/* of clients */
    if (num_clients < MAX_CLIENTS)
      FD_SET(socket, &reads);

/* We select for read on the modem if the serial in buffer is empty */
    if (!serial_in.size)
      FD_SET(modem_in, &reads);

/* If the serial out buffer is empty, try and put something in it */
    if (bytes_left && !serial_out.size) {
      do_serial_out(0);
      if (!serial_out.size)
	do_link_out();
      if (!serial_out.size) 
	do_serial_out(0);
    }

/* Now if there is anything in the serial out buffer, select for writing */

    if (serial_out.size && bytes_left)
      FD_SET(modem_out, &writes);

/* Set the timeout value for select(). */
    timeout.tv_sec = 0;
		/* A nasty bit of code. ;) */
		/* if there is stuff waiting to go out, then there is */
		/* no reason to timeout the select */
    timeout.tv_usec = 500000; /* 0.5 seconds */

/* do select() */	
    if (select(64, &reads, &writes, &excepts, &timeout) < 0) {
      /*	    perror("select"); */
    }	

/* Update current_time. This is maintained in 20th s of a second */
    update_time();

#if 0
/* This shouldn't be nessecery as we trap sig_child to look for */
/* hung children */
    if ((pid = waitpid(0, WNOHANG, 0)) > 0) {
      for (i =0 ; i < MAX_CLIENTS;++i )
	if (clients[i].pid == pid) break;
      if ( i != MAX_CLIENTS )
	clients[i].state = 2;
    }
#endif

/* start checking to see whats ready and whats not */
/* Can we read from modem  ?? */
    if (!serial_in.size && FD_ISSET(modem_in , &reads)) {
      j = read_into_buff(modem_in, &serial_in);
      if ( j < 0 && errno != ERR_BLOCK) {
	perror("read from modem");
	return;
      }
      else if (j > 0)
	stat_modem_in += j;
    }
    
    if (serial_in.size) 
      do_serial_in();

/* Ok. Can we write to modem ??? */
    if (FD_ISSET(modem_out, &writes) && bytes_left) {
      int t = bytes_left;
      if (t > serial_out.size)
	t = serial_out.size;
      j = write_from_buff(modem_out, &serial_out, t);
      if (j < 1 && errno != ERR_BLOCK) {
	perror("write to modem");
	return;
      }	  
      else {
	stat_modem_out += j;
	bytes_left -= j;
      }
    }

/* test for new client */
    if (FD_ISSET(socket, &reads)) { /* try for a connect. */
      struct sockaddr_un dummy;
      int din = sizeof(dummy);
      i = accept(socket , (struct sockaddr *) &dummy, &din);
      if (i >= 0) { /* a new client */
	j = new_client();
	clients[j].fd = i;
	clients[j].cl_type = CL_SOCKET;
	set_nonblock(i);
      }
      else if (errno != ERR_BLOCK) {
	perror("accept");
      }
    }
/* test for data being read from clients */
    for (i = 0; i < MAX_CLIENTS;++i) {
      if (clients[i].fd < 0) continue;
      if (clients[i].in_buff.size) continue;
      switch (clients[i].cl_type) {
      case CL_CHILD:
      case CL_SOCKET:
	if (!FD_ISSET(clients[i].fd, &reads))
	  continue;
      case CL_FILE:
	if (!(clients[i].type & T_RDFILE)) continue;
	j = read_into_buff(clients[i].fd, &clients[i].in_buff); 
	DEBUG_FP(stderr, "%s:read %d bytes from client %d\n",
		 term_server , j, i);
				/* Did this client start sending?? */
	if (j  && j == clients[i].in_buff.size)
	  ++clients_waiting;

				/* Hmmm. Something errored. Lets take */
				/* a look... */
	if (j <= 0)
	  check_client(i, j);
	break;
      case CL_BOUND:
	if (!FD_ISSET(clients[i].fd, &reads)) continue;
	{
	  char num[10]; int k;
				/* We have an accept ready... */
	  clients[i].type &= ~T_RDFILE; /* Don't try reading it untill */
				/* the accept is done. */
	  sprintf(num, "%d", i);
	  for (k =0 ;num[k];++k)
	    add_to_buffer(&clients[i].in_buff, num[k]);
	  add_to_buffer(&clients[i].in_buff, 0);
	}
	break;
      }
    } /* for clients loop */
    
/* test for data being send to clients */
    empty = 1;
    for (i = 0; i < MAX_CLIENTS;++i) {
      if (clients[i].fd < 0) continue;
      if (!clients[i].out_buff.size) continue;
      switch (clients[i].cl_type) {
      case CL_CHILD:
      case CL_SOCKET:
	if  (!FD_ISSET(clients[i].fd, &writes)) continue;
      case CL_FILE:
	if (!(clients[i].type & T_WRFILE)) continue;
				/* something there! :) */
	j = write_from_buff(clients[i].fd, &clients[i].out_buff,0);
	DEBUG_FP(stderr, "%s:write %d bytes to client %d\n", 
		 term_server, j, i);
	if (j <= 0)
	  check_client(i, j);
	break;
      case CL_BOUND:
	break;
      }				/* switch */
    }				/* for clients loop */
    
  } /* while loop */
  
} /* function */

