
#define VERSION	"0.9.9"
#define I_IOCTL
#define I_ERRNO
#define I_SYS
#include "includes.h"

#include <signal.h>
#include <sys/stat.h>
#include <pwd.h>

#ifdef USE_TIOCNOTTY
#include <sys/ioctl.h>
#endif

int pty_pid;

int open_pty(char *prog) {
  int masterfd,
  slavefd,
  c,
  i,
  pid;
  int pip[2];
#ifndef sgi
  char line[]="/dev/ptyXY";
#else
  char *linetmp;
  char line[20];
#endif
  
#ifndef sgi
  /* get a pseudo tty */
  for (c = 'p'; c <= 'z'; ++c) {
    struct stat statbuff;
    
    line[8] = c;
    line[9] = '0';
    
    if (stat (line, &statbuff) < 0)
      return -1; /* no pty's availiable */
    
    for (i = 0; i < 16; ++i) {
      line[9] = "0123456789abcdef"[i];
      if ((masterfd = open (line, O_RDWR)) > 0)
	break;
    }
    if (i != 16)
      break;
  }
  line[5] = 't';
  if ((slavefd = open (line, O_RDWR)) < 0)
    return -2; /* unable to open slave pty */
  
#ifdef SUIDROOT
  if (fchmod (slavefd, 0))
    return -3;
#endif
  
  close(slavefd);

#else /* sgi */
  
  linetmp = _getpty(&masterfd, O_RDWR, 0600, 0);
  
  if (line == NULL)
    return -1; /* no pty's available */
  
  strcpy(line, linetmp);
  
  /* Don't check slave - have to assume it's going to work */
  
#endif /* sgi */
  pipe(pip);
  if ((pid = fork ()) < 0)
    return -3; /* unable to fork */
  
  if (pid == 0) { /* child. */
    char  ** a, *b[2] = {0, };
#if defined(USE_BSDJOBS) || defined(USE_SETPGRP)
    int mypid;
#endif

    close(0);
    close(1);
    close(2);
#ifdef USE_TIOCNOTTY
    if ((slavefd = open ("/dev/tty", O_RDWR)) >= 0) {
      ioctl (slavefd, TIOCNOTTY, (char *) 0);
      close (slavefd);
    }
#endif
#ifdef USE_SETSID
    setsid();
#endif
#ifdef USE_SETPGRP
    mypid = getpid();
    setpgrp(mypid, 0);
#endif
    
#ifdef USE_BSDJOBS
    mypid = getpid();
    tcsetpgrp(0, mypid);
#endif
    
    if ((slavefd = open (line, O_RDWR)) < 0) {
      /* make it control tty */
      exit(1);
    }
#ifdef USE_VHANGUP
    signal (SIGHUP, SIG_IGN);
    vhangup ();
    signal (SIGHUP, SIG_DFL);
#endif
    
    if ((slavefd = open (line, O_RDWR)) < 0) {
      exit(1);
    }
    /*
     **	do a term setup here.
     */
    dup2 (slavefd, 0);
    dup2 (slavefd, 1);
    dup2 (slavefd, 2);
    for (slavefd = 3; slavefd < 64;++slavefd)
      close(slavefd);
    terminal_restore(0);
    /* find out what shell to run. */
    /* note: technically we should free a, but as we are 
     * in the child who will either exec or exit, we can 
     * neglect this.  croutons.
     */
    a = rebuild_arg(prog);
    if ( !a ) {
      a = b; 
    }
    if ( !a[0] )
      a[0] = getenv("SHELL");
    if (!a[0]) a[0] = "/bin/sh";
    printf ("Running on line %s\n", line);
    printf ("Term %s running. about to exec %s...\n", VERSION, a[0]);
    fprintf(stderr, " Execing %s shell\n", a[0]);
    close(pip[0]);
    write(pip[1],"a",1);
    close(pip[1]);
    execvp(a[0], a);
    
    printf ("Unable to exec %s.\n", a[0]);
    exit(1);
  }
				/* Wait for exec.. */
  close(pip[1]);
  read(pip[0], (char *) &i, 1);
  close(pip[0]);

  set_nonblock(masterfd);
  /*	ioctl(masterfd, TIOCPKY, &one); BSD pty packet mode. */
  pty_pid = pid;
  
  return masterfd;
}
