#define I_ERRNO
#define I_SYS
#define I_SOCKET
#include "includes.h"

#include "client.h"

typedef struct {
  int in; 
  int out;
  char buf[1024];
  int size;
} Con;
  
  
int display_num = 9;
int debug = 0;

void main(int argc, char *argv[]) {
  int s;
  int max, first;
  Con *cons = 0;
  int num_cons = 0;

  first = client_options(argc, argv);
  if ( (argc-first) < 1 )
  {
	fprintf ( stderr, "Usage: redir <serverport> <[host:]remoteport>\n" );
	exit(1);
  }
  fprintf ( stderr, "Redirecting %s to %s\n", argv[first], argv[first+1] );

  setbuf(stderr, 0);

				/* Bind the local socket we are going */
				/* to listen on.  */
  s = bind_tcp(atoi(argv[first]));
  if (s < 0) {
    fprintf(stderr, "Port is already bound or some such error."
	    "Exiting\n");
    fprintf(stderr, " s was %d\n", s);
    
    exit(1);
  }

  set_nonblock(s);

  while (1) {
    fd_set in, out;
    int i;
    
    FD_ZERO(&in);
    FD_ZERO(&out);

    FD_SET(s, &in);
    max = s+1;
    for (i = 0; i < num_cons;++i) {
      if (cons[i].in < 0) continue;
      if (!cons[i].size) {
	FD_SET(cons[i].in, &in);
	if (cons[i].in >= max) max = cons[i].in +1;
      }
      else {
	FD_SET(cons[i].out, &out);
	if (cons[i].out >= max) max = cons[i].out + 1;
      }
    }

    select(max, &in, &out, 0, 0);

    if (FD_ISSET(s, &in)) { /* new connection */
      struct sockaddr_in dum;
      int sdum = sizeof(dum),i ;
      for (i = 0; i < num_cons;i += 2)
	if (cons[i].in < 0) break;
      if (i >= num_cons) {
	if (!cons)
	  cons = (Con *) malloc(2 * sizeof(Con));
	else
	  cons = (Con *) realloc((char *) cons, (num_cons + 2) * sizeof(Con));
	i = num_cons;
	num_cons += 2;
      }

      cons[i].in = accept(s, (struct sockaddr *) &dum, &sdum);

      if ((cons[i].out = connect_server(0)) <0) {
	perror("Couldn't open term");
	close(cons[i].in);
	continue;
	}

      if (send_command(cons[i].out, C_PORT, 0, "%s", argv[first+1])< 0) {
	/* An error. Just close off. */
	close(cons[i].in);
	close(cons[i].out);
	cons[i].in = cons[i].out = -1;
	continue;
      }

      send_command(cons[i].out, C_DUMB, 1, 0);
      set_nonblock(cons[i].in);
      set_nonblock(cons[i].out);

      cons[i].size = 0;
      cons[i+1].in = cons[i].out;
      cons[i+1].out = cons[i].in;
      cons[i+1].size =0;
    }

    for (i = 0; i < num_cons;++i) {
      if (cons[i].in < 0) continue;
      if (!FD_ISSET(cons[i].in, &in)) continue;
      cons[i].size = read(cons[i].in, cons[i].buf, 1024);
/* Handle possible error condition */
      if (cons[i].size < 0 && errno == ERR_BLOCK) {
	/* Some brain dead OS's need this and it doesn't hurt */
	cons[i].size = 0;
	continue;
      }
      if (cons[i].size <= 0) {
	/* an error has occured or a stream has closed */
	/* Close connection. NYF */
	i &= ~1; /* Round down to even number */
	close(cons[i].out);
	close(cons[i].in);
	cons[i].in = cons[i].out = -1;
	cons[i+1].in = cons[i+1].out = -1;
	continue;
      }
    }

    for (i = 0; i < num_cons;++i) {
      int j;
      if (cons[i].out < 0) continue;
      if (!FD_ISSET(cons[i].out, &out)) continue;
      j = write(cons[i].out, cons[i].buf, cons[i].size);
/* Handle possible error condition */
      if (j <0 && errno== ERR_BLOCK ) {
	cons[i].size = 0;
	continue;
      }
      if (j <=0) {
	i &= ~1;
	close(cons[i].in);
	close(cons[i].out);
	cons[i].in = cons[i].out = -1;
	cons[i+1].in = cons[i+1].out = -1;
	continue;
      }

      if (j == cons[i].size) {
	cons[i].size = 0;
	continue;
      }

      memcpy(cons[i].buf, cons[i].buf + j, cons[i].size - j);
      cons[i].size -= j;
    }
     
  }
  set_block(s);
}

