#define I_ERRNO
#define I_SYS
#define I_IOCTL
#define I_STRING
#define I_SOCKET

#include "includes.h"

#include "client.h"

typedef struct {
  int in; 
  int out;
  char buf[1024];
  int size;
} Con;

#ifdef X_DEBUG
int d_fd= -1, sock;
#endif
  
char *local_x =  "/tmp/.X11-unix/X0";

int display_num = 9;
int debug = 0;

void main(int argc, char *argv[]) {
  int s;
  char *home;
  int max;
  Con *cons = 0;
  int num_cons = 0;
  char *display;
#ifdef UNIX_SOCKET
  char unix_sock[110];
#endif

  setbuf(stderr, 0);
  display = getenv("DISPLAY");

#ifdef UNIX_SOCKET
  if (!display || display[0] != ':' || !display[1])
    display = ":9";

  sprintf(unix_sock,"/tmp/.X11-unix");
  mkdir(unix_sock, 0700);
  strcat(unix_sock, "/X");
  strcat(unix_sock, display+1);

  if ((s = bind_unix(unix_sock)) < 0) {
    exit(1);
  }

#else /* if a TCP socket.. */
				/* Try and work out what display we */
				/* should get by default. */
  if (!display || !display[0]) { /* is there a DISPLAY set?? */
    display_num = 9;		/* No. Just pick 9 by default */
  } else {
    home = strchr(display,':');	/* Yes. Find out what display number. */
    if (home)			/* if we could work it out.. */
      display_num = atoi(home+1); /* then grab the number. */
    if (display_num < 1)	/* if the number is invalid,  */
      display_num = 9;		/* just default. */
  }
  
  while (( s = bind_tcp(6000 + display_num)) == -2) {
    if (display_num > 100) {
      printf("Unable to bind socket\n");
      exit(1);
      break;
    }
    ++display_num;
  }

#endif

  if (fork()) {
    exit(0);
  }


  set_nonblock(s);

  while (1) {
    fd_set in, out;
    int i;
    
    FD_ZERO(&in);
    FD_ZERO(&out);

    FD_SET(s, &in);
    max = s+1;
    for (i = 0; i < num_cons;++i) {
      if (cons[i].in < 0) continue;
      if (!cons[i].size) {
	FD_SET(cons[i].in, &in);
	if (cons[i].in >= max) max = cons[i].in +1;
      }
      else {
	FD_SET(cons[i].out, &out);
	if (cons[i].out >= max) max = cons[i].out + 1;
      }
    }

    select(max, &in, &out, 0, 0);

    if (FD_ISSET(s, &in)) { /* new connection */
#ifdef UNIX_SOCKET
      struct sockaddr_un dum;
#else
      struct sockaddr_in dum;
#endif
      int sdum = sizeof(dum),i ;
      for (i = 0; i < num_cons;i += 2)
	if (cons[i].in < 0) break;
      if (i >= num_cons) {
	if (!cons)
	  cons = (Con *) malloc(2 * sizeof(Con));
	else
	  cons = (Con *) realloc((char *) cons, (num_cons + 2) * sizeof(Con));
	i = num_cons;
	num_cons += 2;
      }

      cons[i].in = accept(s, (struct sockaddr *) &dum, &sdum);

      if ((cons[i].out = connect_server(0)) <0) {
	perror("Couldn't open term");
	close(cons[i].in);
	continue;
	}

      if (send_command(cons[i].out, C_SOCKET, 0, "%s", local_x)<0) {
	close(cons[i].out);
	close(cons[i].in);
	cons[i].out = cons[i].in = -1;
	continue;
      }
      send_command(cons[i].out, C_DUMB, 1, 0);
#ifdef X_DEBUG
      if (d_fd <0) {
	d_fd = open("x.debug", O_RDWR | O_CREAT | O_TRUNC);
	sock = cons[i].out;
      }
#endif
      set_nonblock(cons[i].in);
      set_nonblock(cons[i].out);

      cons[i].size = 0;
      cons[i+1].in = cons[i].out;
      cons[i+1].out = cons[i].in;
      cons[i+1].size =0;
    }

    for (i = 0; i < num_cons;++i) {
      if (cons[i].in < 0) continue;
      if (!FD_ISSET(cons[i].in, &in)) continue;
      cons[i].size = read(cons[i].in, cons[i].buf, 1024);
      /* Handle possible error condition */
      if (cons[i].size < 0 && errno == ERR_BLOCK) {
	/* Some brain dead OS's need this and it doesn't hurt */
	cons[i].size = 0;
	continue;
      }
      if (cons[i].size <= 0) {
	/* an error has occured or a stream has closed */
	/* Close connection. NYF */
	i &= ~1; /* Round down to even number */
#ifdef X_DEBUG
	if (cons[i].out == sock)
	  close(d_fd);
#endif
	close(cons[i].out);
	close(cons[i].in);
	cons[i].in = cons[i].out = -1;
	cons[i+1].in = cons[i+1].out = -1;
	continue;
      }
    }

    for (i = 0; i < num_cons;++i) {
      int j;
      if (cons[i].out < 0) continue;
      if (!FD_ISSET(cons[i].out, &out)) continue;
#ifdef X_DEBUG
      if (cons[i].out == sock)
	write(d_fd, cons[i].buf, cons[i].size);
#endif
      j = write(cons[i].out, cons[i].buf, cons[i].size);
/* Handle possible error condition */
      if (j <0 && errno== ERR_BLOCK ) {
	cons[i].size = 0;
	continue;
      }
      if (j <=0) {
	i &= ~1;
	close(cons[i].in);
	close(cons[i].out);
	cons[i].in = cons[i].out = -1;
	cons[i+1].in = cons[i+1].out = -1;
	continue;
      }

      if (j == cons[i].size) {
	cons[i].size = 0;
	continue;
      }

      memcpy(cons[i].buf, cons[i].buf + j, cons[i].size - j);
      cons[i].size -= j;
    }
     
  }
  set_block(s);
}

